import json
import logging
import tarfile
from io import BytesIO
from pathlib import Path
from tempfile import NamedTemporaryFile
from unittest.mock import patch
from zipfile import ZipFile

from zstandard import ZstdCompressor

SAMPLES = Path(__file__).absolute().parent / "samples"


@patch("worker_file_import.worker.TranskribusImporter")
def test_run_zip(importer_mock, mock_worker, responses, caplog):
    caplog.set_level(logging.ERROR)

    # Process info
    mock_worker.process_information = {
        "id": "processid",
        "corpus": "corpusid",
        "mode": "files",
        "element": {
            "id": "folderid",
            "name": "Untitled",
            "type": "book",
            "corpus": {"id": "corpusid"},
        },
        "files": ["file1", "file2", "file3", "file4"],
        "element_type": "page",
        "folder_type": "book",
    }
    # DataFile info
    mock_worker.api_client.add_response(
        "RetrieveDataFile",
        id="file1",
        response={
            "id": "file1",
            "name": "file1.zip",
            "content_type": "application/zip",
            "s3_url": "http://s3/file1.zip",
        },
    )
    mock_worker.api_client.add_response(
        "RetrieveDataFile",
        id="file2",
        response={
            "id": "file2",
            "name": "file2.zip",
            "content_type": "application/zip",
            "s3_url": "http://s3/file2.zip",
        },
    )
    mock_worker.api_client.add_response(
        "RetrieveDataFile",
        id="file3",
        response={
            "id": "file3",
            "name": "file3.zip",
            # Use the Windows MIME type
            "content_type": "application/x-zip-compressed",
            "s3_url": "http://s3/file3.zip",
        },
    )
    mock_worker.api_client.add_response(
        "RetrieveDataFile",
        id="file4",
        response={
            "id": "file4",
            "name": "file4.zip",
            "content_type": "application/zip",
            "s3_url": "http://s3/file4.zip",
        },
    )

    # Create an archive that looks like a Transkribus export
    transkribus_zip = NamedTemporaryFile()
    with ZipFile(transkribus_zip, "a") as archive:
        archive.writestr("1234/document_title/mets.xml", "lol")
    transkribus_zip.flush()
    transkribus_zip.seek(0)
    zip_path = Path(transkribus_zip.name)
    responses.add(responses.GET, "http://s3/file1.zip", body=zip_path.open("rb"))

    # An archive that isn't a Transkribus export
    dummy_zip = NamedTemporaryFile()
    with ZipFile(dummy_zip, "a") as archive:
        archive.writestr("lol.txt", "lol")
    dummy_zip.flush()
    dummy_zip.seek(0)
    zip_path = Path(dummy_zip.name)
    responses.add(responses.GET, "http://s3/file2.zip", body=zip_path.open("rb"))

    # Windows MIME type test: we just use a dummy archive.
    # We cannot reuse the dummy_zip defined above since it will be consumed by the mocked request
    windows_zip = NamedTemporaryFile()
    with ZipFile(windows_zip, "a") as archive:
        archive.writestr("lol.txt", "lol")
    windows_zip.flush()
    windows_zip.seek(0)
    zip_path = Path(windows_zip.name)
    responses.add(responses.GET, "http://s3/file3.zip", body=zip_path.open("rb"))

    # Not actually a zip file
    responses.add(
        responses.GET, "http://s3/file4.zip", body=(SAMPLES / "file2.pdf").open("rb")
    )

    for filename in ["file2", "file3", "file4"]:
        mock_worker.api_client.add_response(
            "CreateElement",
            body={
                "corpus": "corpusid",
                "type": "book",
                "name": f"{filename}.zip",
                "parent": "folderid",
                "worker_run_id": mock_worker.worker_run_id,
            },
            response={"id": f"id_{filename}", "type": "book"},
        )

    # Process update
    mock_worker.api_client.add_response(
        "PartialUpdateProcess",
        id="processid",
        body={"element_id": "folderid"},
        response={},
    )
    # DataFile deletion
    mock_worker.api_client.add_response("DestroyDataFile", id="file1", response={})

    importer_mock.run.return_value = [{"id": "imported", "type": "thing"}]

    mock_worker.run()

    # file1.zip triggers a Transkribus import
    importer_mock.assert_called_once()

    # Two position arguments: the worker and a path to the temporary file
    # that the file import uses to download the file.
    assert len(importer_mock.call_args.args) == 2
    # To check that this is the Transkribus export, we open the archive and look for the XML file we put in there
    with ZipFile(importer_mock.call_args.args[1]) as archive:
        assert archive.namelist() == ["1234/document_title/mets.xml"]

    assert importer_mock.call_args.kwargs == {
        "parent_element": {
            "id": "folderid",
            "name": "Untitled",
            "type": "book",
            "corpus": {"id": "corpusid"},
        },
    }

    assert caplog.record_tuples == [
        # Dummy ZIP file
        (
            "worker_file_import.worker",
            40,
            "Could not import file lol.txt from archive file2.zip: NotImplementedError: ",
        ),
        (
            "worker_file_import.worker",
            40,
            "Could not import file file2.zip: AssertionError: Not all files from archive file2.zip have been imported.",
        ),
        # Dummy ZIP file with Windows MIME type
        (
            "worker_file_import.worker",
            40,
            "Could not import file lol.txt from archive file3.zip: NotImplementedError: ",
        ),
        (
            "worker_file_import.worker",
            40,
            "Could not import file file3.zip: AssertionError: Not all files from archive file3.zip have been imported.",
        ),
        # Not a ZIP file
        (
            "worker_file_import.transkribus",
            40,
            "File file4.zip does not appear to be a valid ZIP archive.",
        ),
        (
            "worker_file_import.worker",
            40,
            "Could not import file file4.zip: BadZipFile: File is not a zip file",
        ),
    ]

    with (mock_worker.work_dir / "elements.json").open() as f:
        elements = json.load(f)
    assert elements == [
        {"id": "folderid", "type": "book"},
        {"id": "id_file2", "type": "book"},
        {"id": "id_file3", "type": "book"},
        {"id": "id_file4", "type": "book"},
    ]


def test_run_tar(mock_worker, responses, caplog):
    caplog.set_level(logging.ERROR)

    # Prepare all of the archives to test
    cases = [
        ("bzip", "application/x-bzip", "w:bz2"),
        ("bzip2", "application/x-bzip2", "w:bz2"),
        ("gz", "application/gzip", "w:gz"),
        ("lzma", "application/x-lzma", "w:xz"),
        ("tar", "application/x-tar", "w:"),
        ("tar-gz", "application/x-gtar", "w:gz"),
        ("xz", "application/x-xz", "w:xz"),
    ]

    # tarfile does not support an equivalent of ZipFile.writestr, so we need an actual file to add to archives.
    with NamedTemporaryFile() as text_file:
        text_file.write(b"lol")
        text_file.flush()

        for file_id, _, mode in cases:
            archive_file = NamedTemporaryFile()
            with tarfile.open(fileobj=archive_file, mode=mode) as archive:
                archive.add(text_file.name, arcname="lol.txt")
            archive_file.flush()
            archive_file.seek(0)
            tar_path = Path(archive_file.name)
            responses.add(
                responses.GET, f"http://s3/{file_id}", body=tar_path.open("rb")
            )

        # .tar.zst is handled separately, since tarfile doesn't support it
        with NamedTemporaryFile() as archive_file:
            with tarfile.open(fileobj=archive_file, mode="w:") as archive:
                archive.add(text_file.name, arcname="lol.txt")
            archive_file.flush()
            archive_file.seek(0)

            zst_archive = NamedTemporaryFile()
            ZstdCompressor().copy_stream(archive_file, zst_archive)
            zst_archive.seek(0)
            zst_path = Path(zst_archive.name)
            responses.add(responses.GET, "http://s3/zst", body=zst_path.open("rb"))

        cases.append(("zst", "application/zstd", None))

    # Build RetrieveDataFile mocks
    for file_id, content_type, _ in cases:
        mock_worker.api_client.add_response(
            "RetrieveDataFile",
            id=file_id,
            response={
                "id": file_id,
                "name": file_id,
                "content_type": content_type,
                "s3_url": f"http://s3/{file_id}",
            },
        )

    # Process info
    mock_worker.process_information = {
        "id": "processid",
        "corpus": "corpusid",
        "mode": "files",
        "element": {
            "id": "folderid",
            "name": "Untitled",
            "type": "book",
            "corpus": {"id": "corpusid"},
        },
        "files": [file_id for file_id, _, _ in cases],
        "element_type": "page",
        "folder_type": "book",
    }

    # Mock all folder creations
    for file_id, _, _ in cases:
        mock_worker.api_client.add_response(
            "CreateElement",
            body={
                "corpus": "corpusid",
                "type": "book",
                "name": file_id,
                "parent": "folderid",
                "worker_run_id": mock_worker.worker_run_id,
            },
            response={"id": "newfolderid", "type": "book"},
        )

    # Mock PartialUpdateProcess to allow setting Process.element
    mock_worker.api_client.add_response(
        "PartialUpdateProcess",
        id="processid",
        body={"element_id": "folderid"},
        response={},
    )

    mock_worker.run()

    for file_id, _, _ in cases:
        responses.assert_call_count(f"http://s3/{file_id}", 1)

    expected_logs = []
    for file_id, _, _ in cases:
        expected_logs.extend(
            [
                (
                    "worker_file_import.worker",
                    40,
                    f"Could not import file lol.txt from archive {file_id}: NotImplementedError: ",
                ),
                (
                    "worker_file_import.worker",
                    40,
                    f"Could not import file {file_id}: AssertionError: Not all files from archive {file_id} have been imported.",
                ),
            ]
        )

    assert caplog.record_tuples == expected_logs


def test_zipception(mock_worker, responses, caplog):
    caplog.set_level(logging.ERROR)

    zip4 = NamedTemporaryFile()
    with BytesIO() as zip1, BytesIO() as zip2, BytesIO() as zip3:
        with ZipFile(zip1, "w") as archive:
            archive.writestr("lol.txt", "lol")

        with ZipFile(zip2, "w") as archive:
            archive.writestr("lol.zip", zip1.getvalue())

        with ZipFile(zip3, "w") as archive:
            archive.writestr("lol.zip.zip", zip2.getvalue())

        with ZipFile(zip4, "w") as archive:
            archive.writestr("lol.zip.zip.zip", zip3.getvalue())
    zip4.flush()
    zip4.seek(0)
    zip_path = Path(zip4.name)
    responses.add(responses.GET, "http://s3/file1.zip", body=zip_path.open("rb"))

    # Process info
    mock_worker.process_information = {
        "id": "processid",
        "corpus": "corpusid",
        "mode": "files",
        "element": {
            "id": "folderid",
            "name": "Untitled",
            "type": "book",
            "corpus": {"id": "corpusid"},
        },
        "files": ["file1"],
        "element_type": "page",
        "folder_type": "book",
    }

    # DataFile info
    mock_worker.api_client.add_response(
        "RetrieveDataFile",
        id="file1",
        response={
            "id": "file1",
            "name": "file1.zip",
            "content_type": "application/zip",
            "s3_url": "http://s3/file1.zip",
        },
    )

    # Create folder responses
    for filename in ["file1.zip", "lol.zip.zip.zip", "lol.zip.zip"]:
        mock_worker.api_client.add_response(
            "CreateElement",
            body={
                "corpus": "corpusid",
                "type": "book",
                "name": filename,
                "parent": "folderid",
                "worker_run_id": mock_worker.worker_run_id,
            },
            response={"id": "folderid", "type": "book"},
        )

    # Process update
    mock_worker.api_client.add_response(
        "PartialUpdateProcess",
        id="processid",
        body={"element_id": "folderid"},
        response={},
    )

    mock_worker.run()

    assert caplog.record_tuples == [
        (
            "worker_file_import.worker",
            40,
            "Could not import file lol.zip from archive lol.zip.zip: AssertionError: Maximum archive decompression depth reached",
        ),
        (
            "worker_file_import.worker",
            40,
            "Could not import file lol.zip.zip from archive lol.zip.zip.zip: AssertionError: Not all files from archive lol.zip.zip have been imported.",
        ),
        (
            "worker_file_import.worker",
            40,
            "Could not import file lol.zip.zip.zip from archive file1.zip: AssertionError: Not all files from archive lol.zip.zip.zip have been imported.",
        ),
        (
            "worker_file_import.worker",
            40,
            "Could not import file file1.zip: AssertionError: Not all files from archive file1.zip have been imported.",
        ),
    ]

    with (mock_worker.work_dir / "elements.json").open() as f:
        elements = json.load(f)
    assert elements == [
        {"id": "folderid", "type": "book"},
        {"id": "folderid", "type": "book"},
        {"id": "folderid", "type": "book"},
        {"id": "folderid", "type": "book"},
    ]
