import json
from io import BufferedReader
from pathlib import Path
from unittest.mock import call, patch

import pytest

from arkindex_worker.models import Element
from worker_file_import.iiif import to_primitive, trim_warn

SAMPLES = Path(__file__).absolute().parent / "samples"


@patch("worker_file_import.worker.IIIFParser")
def test_run_iiif(parser_mock, mock_worker, responses):
    # Process info
    mock_worker.process_information = {
        "id": "processid",
        "corpus": "corpusid",
        "mode": "files",
        "element": {
            "id": "folderid",
            "name": "Untitled",
            "type": "potato",
            "corpus": {"id": "corpusid"},
        },
        "files": ["file1", "file2"],
        "folder_type": "potato",
        "element_type": "page",
    }

    # DataFile info
    mock_worker.api_client.add_response(
        "RetrieveDataFile",
        id="file1",
        response={
            "id": "file1",
            "name": "manifest.json",
            "s3_url": "http://somewhere.com/IIIF/file1/manifest.json",
            "content_type": "application/json",
        },
    )
    mock_worker.api_client.add_response(
        "RetrieveDataFile",
        id="file2",
        response={
            "id": "file2",
            "name": "collection.json",
            "s3_url": "http://somewhere.com/IIIF/file2/collection.json",
            "content_type": "application/json",
        },
    )

    # DataFile on S3
    responses.add(
        responses.GET,
        "http://somewhere.com/IIIF/file1/manifest.json",
        body=(SAMPLES / "manifest.json").open("rb"),
    )
    responses.add(
        responses.GET,
        "http://somewhere.com/IIIF/file2/collection.json",
        body=(SAMPLES / "collection.json").open("rb"),
    )

    # Process update
    mock_worker.api_client.add_response(
        "PartialUpdateProcess",
        id="processid",
        body={"element_id": "folderid"},
        response={},
    )

    # DataFile deletion
    mock_worker.api_client.add_response("DestroyDataFile", id="file1", response={})
    mock_worker.api_client.add_response("DestroyDataFile", id="file2", response={})

    parser_mock.load.return_value.elements = [{"id": "elementid", "type": "page"}]

    mock_worker.run()

    call1, call2 = parser_mock.load.call_args_list
    args, kwargs = call1
    assert len(args) == 2
    assert isinstance(args[1], BufferedReader)
    assert kwargs == {
        "parent_element": Element(
            {
                "id": "folderid",
                "name": "Untitled",
                "type": "potato",
                "corpus": {"id": "corpusid"},
            }
        )
    }
    args, kwargs = call2
    assert len(args) == 2
    assert isinstance(args[1], BufferedReader)
    assert kwargs == {
        "parent_element": Element(
            {
                "id": "folderid",
                "name": "Untitled",
                "type": "potato",
                "corpus": {"id": "corpusid"},
            }
        )
    }

    elements = mock_worker.work_dir / "elements.json"
    assert elements.is_file()
    with elements.open() as f:
        assert json.load(f) == [
            {"id": "folderid", "type": "potato"},
            {"id": "elementid", "type": "page"},
            {"id": "elementid", "type": "page"},
        ]

    responses.assert_call_count("http://somewhere.com/IIIF/file2/collection.json", 1)
    responses.assert_call_count("http://somewhere.com/IIIF/file1/manifest.json", 1)


# Helpers


def test_primitive_to_primitive():
    assert to_primitive("lol") == "lol"
    assert to_primitive(42) == 42
    assert to_primitive(True)
    assert to_primitive(None) is None


def test_dict_to_primitive():
    assert to_primitive({"a": "b", "@value": "something"}) == "something"


def test_list_to_primitive():
    assert (
        to_primitive(
            [
                {"a": "b", "@value": "something"},
                {"c": "d", "@value": "somethingelse"},
            ]
        )
        == "something"
    )
    assert (
        to_primitive(
            [
                {"a": "b", "@value": "something"},
                {"c": "d", "@language": "en-US", "@value": "somethingelse"},
            ]
        )
        == "somethingelse"
    )


def test_invalid_objects_to_primitive():
    with pytest.raises(
        ValueError,
        match="Expected primitive or JSON-LD Value Object, got an object without a @value property",
    ):
        to_primitive({"a": "b"})

    with pytest.raises(
        ValueError,
        match="Expected primitive or JSON-LD Value Object, got an object without a @value property",
    ):
        to_primitive([{"a": "b"}, {"c": "d", "@value": "something"}])

    with pytest.raises(
        ValueError, match="Array of JSON-LD Value Objects should not be empty"
    ):
        to_primitive([])

    with pytest.raises(
        ValueError, match="Array should only hold JSON-LD Value Objects"
    ):
        to_primitive([{"@value": "something"}, 42, True])


@patch("worker_file_import.iiif.logger")
def test_trim_warn(logger_mock):
    assert trim_warn("A" * 42, name="Something", max_length=42) == "A" * 42
    assert logger_mock.warning.call_count == 0

    assert trim_warn("A" * 1337, name="Something", max_length=42) == "A" * 42
    assert logger_mock.warning.call_count == 1
    assert logger_mock.warning.call_args == call(
        "Something is longer than the 42 character limit and will be trimmed"
    )


def test_many():
    assert to_primitive("a", many=True) == ["a"]
    assert to_primitive(["a", "b"], many=True) == ["a", "b"]
    assert to_primitive([{"@value": "a"}, {"@value": "b"}], many=True) == ["a", "b"]
    assert to_primitive(
        [{"@value": "a"}, {"@value": "b", "@language": "cn"}], many=True
    ) == ["a", "b"]
    assert to_primitive(
        [
            {"@value": "a", "@language": "en"},
            {"@value": "b", "@language": "cn"},
        ],
        many=True,
    ) == ["a"]
