import tempfile
from logging import Logger, getLogger
from pathlib import Path

import requests
from apistar.exceptions import ErrorResponse
from PIL import Image, ImageOps
from teklia_toolbox.requests import should_verify_cert
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_fixed

from worker_file_import.utils import hash_image

logger: Logger = getLogger(__name__)

# Image formats that should not be converted
# see https://pillow.readthedocs.io/en/5.1.x/handbook/image-file-formats.html
VALID_FORMATS = ["JPEG", "JPEG2000", "PNG"]


def check_image(datafile):
    try:
        img = Image.open(datafile["local_path"])
    except OSError as err:
        raise ValueError(
            "File {} is not a valid image".format(datafile["name"])
        ) from err

    # Keep track of whether or not the image needs to be re-saved
    changed = False

    img_path = Path(tempfile.mkdtemp()) / datafile["id"]
    img_format = img.format

    # Default to JPEG
    if img_format not in VALID_FORMATS:
        img_path = f"{img_path}.jpg"
        img_format = "JPEG"
        changed = True

    if img.getexif().get(0x0112, 0) not in (0, 1):
        logger.info(f"Image ({img.size}) is rotated, it will be transposed.")
        # Apply the conversion specified by the EXIF Orientation tag if it is set and is not 1
        # (1 is TopLeft, which is equivalent to doing nothing)
        img = ImageOps.exif_transpose(img)
        changed = True
        assert img.getexif().get(0x0112, 0) in (
            0,
            1,
        ), "Image should not be rotated anymore"

    # Convert to RGB if needed
    if img.mode != "RGB":
        img = img.convert("RGB")
        changed = True

    if changed:
        img.save(img_path, format=img_format)
        return [Path(img_path)]
    return [datafile["local_path"]]


@retry(
    reraise=True,
    retry=retry_if_exception_type(requests.RequestException),
    stop=stop_after_attempt(3),
    wait=wait_fixed(5),
)
def upload_image(path, url):
    with Path.open(path, "rb") as f:
        resp = requests.put(url, data=f, verify=should_verify_cert(url))
    resp.raise_for_status()


def create_image(api_client, path):
    logger.debug(f"Creating image {path}")
    hash = hash_image(path)

    try:
        image = api_client.request("CreateImage", body={"hash": hash})
    except ErrorResponse as e:
        if e.status_code == 400 and "id" in e.content:
            image = api_client.request("RetrieveImage", id=e.content["id"])
        else:
            raise
    else:
        logger.info("Uploading image {}".format(image["id"]))
        upload_image(path, image["s3_put_url"])

    logger.debug("Validating image {}".format(image["id"]))
    api_client.request("PartialUpdateImage", id=image["id"], body={"status": "checked"})

    return image["id"]
